/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */ 
package com.rimfaxe.webserver.compiler.jsp;

import java.util.Hashtable;
import java.util.Enumeration;
import javax.servlet.jsp.tagext.PageData;
import javax.servlet.jsp.tagext.TagData;
import javax.servlet.jsp.tagext.TagInfo;
import javax.servlet.jsp.tagext.TagAttributeInfo;
import javax.servlet.jsp.tagext.TagLibraryInfo;
import com.rimfaxe.webserver.webapp.TagLibraryInfoImpl;
import javax.servlet.jsp.tagext.ValidationMessage;

import org.xml.sax.Attributes;
import com.rimfaxe.webserver.compiler.JspToJavaException;

/**
 * Performs validation on the page elements.  Attributes are checked for
 * mandatory presence, entry value validity, and consistency.  As a
 * side effect, some page global value (such as those from page direcitves)
 * are stored, for later use.
 *
 * @author Kin-man Chung
 * @author Jan Luehe
 * @author Lars Andersen
 */
public class Validator 
{

    /**
     * A visitor to validate and extract page directive info
     */
    static class PageDirectiveVisitor extends Node.Visitor 
    {

	private PageInfo pageInfo;
        //private JspErrorHandler err;
        

	private static final JspUtil.ValidAttribute[] pageDirectiveAttrs = 
        {
	    new JspUtil.ValidAttribute("language"),
	    new JspUtil.ValidAttribute("extends"),
	    new JspUtil.ValidAttribute("import"),
	    new JspUtil.ValidAttribute("session"),
	    new JspUtil.ValidAttribute("buffer"),
	    new JspUtil.ValidAttribute("autoFlush"),
	    new JspUtil.ValidAttribute("isThreadSafe"),
	    new JspUtil.ValidAttribute("info"),
	    new JspUtil.ValidAttribute("errorPage"),
	    new JspUtil.ValidAttribute("isErrorPage"),
	    new JspUtil.ValidAttribute("contentType"),
	    new JspUtil.ValidAttribute("pageEncoding") };

	private boolean languageSeen = false;
	private boolean extendsSeen = false;
	private boolean sessionSeen = false;
 	private boolean bufferSeen = false;
	private boolean autoFlushSeen = false;
	private boolean isThreadSafeSeen = false;
	private boolean errorPageSeen = false;
	private boolean isErrorPageSeen = false;
	private boolean contentTypeSeen = false;
	private boolean infoSeen = false;
	private boolean pageEncodingSeen = false;

	/*
	 * Constructor
	 */
	PageDirectiveVisitor(JspC compiler) 
        {
	    this.pageInfo = compiler.getPageInfo();
	    //this.err = compiler.getErrorHandler();
	}

	public void visit(Node.PageDirective n)  
            throws JasperException, JspToJavaException  
        {    

	    JspUtil.checkAttributes("Page directive", n.getAttributes(), pageDirectiveAttrs, n.getStart());

	    // JSP.2.10.1
	    Attributes attrs = n.getAttributes();
	    for (int i = 0; i < attrs.getLength(); i++) {
		String attr = attrs.getQName(i);
		String value = attrs.getValue(i);

		if ("language".equals(attr)) 
                {
		    if (languageSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple languages");
		    	
		    languageSeen = true;
		    if (!"java".equalsIgnoreCase(value))
		    	throw new JspToJavaException(n.getStart(), "Nonjava language");
		    pageInfo.setLanguage(value);
		} 
                else if ("extends".equals(attr)) 
                {
		    if (extendsSeen)
		    	throw new JspToJavaException(n.getStart(), "Multiple extends");
		    extendsSeen = true;
		    pageInfo.setExtends(value);
		    /*
		     * If page superclass is top level class (i.e. not in a
		     * pkg) explicitly import it. If this is not done, the
		     * compiler will assume the extended class is in the same
		     * pkg as the generated servlet.
		     */
		    if (value.indexOf('.') < 0)
			n.addImport(value);
		} 
                else if ("contentType".equals(attr)) 
                {
		    if (contentTypeSeen) 
                        throw new JspToJavaException(n.getStart(), "Multiple contenttypes");
		    	
		    contentTypeSeen = true;
		    pageInfo.setContentType(value);
		} 
                else if ("session".equals(attr)) 
                {
		    if (sessionSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple sessions");
		    	
		    sessionSeen = true;
		    if ("true".equalsIgnoreCase(value))
			pageInfo.setSession(true);
		    else if ("false".equalsIgnoreCase(value))
			pageInfo.setSession(false);
		    else
                        throw new JspToJavaException(n.getStart(), "Invalid session");
                        
			
		} 
                else if ("buffer".equals(attr)) 
                {
		    if (bufferSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple buffers");
		    	
		    bufferSeen = true;

		    if ("none".equalsIgnoreCase(value))
			pageInfo.setBuffer(0);
		    else 
                    {
			if (value == null || !value.endsWith("kb"))
                            throw new JspToJavaException(n.getStart(), "Invalid buffer");
			    

			try 
                        {
			    Integer k = new Integer(value.substring(0, value.length()-2));
			    pageInfo.setBuffer(k.intValue()*1024);
			} 
                        catch (NumberFormatException e) 
                        {
                           System.out.println("NumberFormatException in Validator"); 
                           throw new JspToJavaException(n.getStart(), "Invalid buffer"); 
			  
			}
		    }
		} 
                else if ("autoFlush".equals(attr)) 
                {
		    if (autoFlushSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple autoflush");
		    	
		    autoFlushSeen = true;
		    if ("true".equalsIgnoreCase(value))
			pageInfo.setAutoFlush(true);
		    else if ("false".equalsIgnoreCase(value))
			pageInfo.setAutoFlush(false);
		    else
                        throw new JspToJavaException(n.getStart(), "Autoflush invalid");
		    	
		} 
                else if ("isthreadSafe".equals(attr)) 
                {
		    if (isThreadSafeSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple isThreadsafe");
			
		    isThreadSafeSeen = true;
		    if ("true".equalsIgnoreCase(value))
			pageInfo.setThreadSafe(true);
		    else if ("false".equalsIgnoreCase(value))
			pageInfo.setThreadSafe(false);
		    else
                        throw new JspToJavaException(n.getStart(), "isThreadSafe invalid");
			
		} 
                else if ("isErrorPage".equals(attr)) 
                {
		    if (isErrorPageSeen)
                        throw new JspToJavaException(n.getStart(), "Multiple isErrorPage");
			
		    isErrorPageSeen = true;
		    if ("true".equalsIgnoreCase(value))
			pageInfo.setIsErrorPage(true);
		    else if ("false".equalsIgnoreCase(value))
			pageInfo.setIsErrorPage(false);
		    else
                        throw new JspToJavaException(n.getStart(), "isErrorPage invalid");
			
		} 
                else if ("errorPage".equals(attr)) 
                {
		    if (errorPageSeen) 
                        throw new JspToJavaException(n.getStart(), "Multiple errorpage");
			
		    errorPageSeen = true;
		    pageInfo.setErrorPage(value);
		} 
                else if ("info".equals(attr)) 
                {
		    if (infoSeen) 
                        throw new JspToJavaException(n.getStart(), "Multiple info");
			
		    infoSeen = true;
		} 
                else if ("pageEncoding".equals(attr)) 
                {
		    if (pageEncodingSeen) 
                        throw new JspToJavaException(n.getStart(), "Multiple page encodings");
			
		    pageEncodingSeen = true;
		    pageInfo.setPageEncoding(value);
		}
	    }

	    // Check for bad combinations
	    if (pageInfo.getBuffer() == 0 && !pageInfo.isAutoFlush())
                throw new JspToJavaException(n.getStart(), "Bad combo");
		

	    // Attributes for imports for this node have been processed by
	    // the parsers, just add them to pageInfo.
	    pageInfo.addImports(n.getImports());
	}
    }

    /**
     * A visitor for validating nodes other than page directives
     */
    static class ValidateVisitor extends Node.Visitor 
    {

	private PageInfo pageInfo;
	

	private static final JspUtil.ValidAttribute[] jspRootAttrs = 
        {
	    new JspUtil.ValidAttribute("version", true) 
        };

	private static final JspUtil.ValidAttribute[] includeDirectiveAttrs = {
	    new JspUtil.ValidAttribute("file", true) };

	private static final JspUtil.ValidAttribute[] taglibDirectiveAttrs = {
	    new JspUtil.ValidAttribute("uri", true),
	    new JspUtil.ValidAttribute("prefix", true) };

	private static final JspUtil.ValidAttribute[] includeActionAttrs = {
	    new JspUtil.ValidAttribute("page", true),
	    new JspUtil.ValidAttribute("flush") };

	private static final JspUtil.ValidAttribute[] paramActionAttrs = {
	    new JspUtil.ValidAttribute("name", true),
	    new JspUtil.ValidAttribute("value", true) };

	private static final JspUtil.ValidAttribute[] forwardActionAttrs = {
	    new JspUtil.ValidAttribute("page", true) };

	private static final JspUtil.ValidAttribute[] getPropertyAttrs = {
	    new JspUtil.ValidAttribute("name", true),
	    new JspUtil.ValidAttribute("property", true) };

	private static final JspUtil.ValidAttribute[] setPropertyAttrs = {
	    new JspUtil.ValidAttribute("name", true),
	    new JspUtil.ValidAttribute("property", true),
	    new JspUtil.ValidAttribute("value"),
	    new JspUtil.ValidAttribute("param") };

	private static final JspUtil.ValidAttribute[] useBeanAttrs = {
	    new JspUtil.ValidAttribute("id", true),
	    new JspUtil.ValidAttribute("scope"),
	    new JspUtil.ValidAttribute("class"),
	    new JspUtil.ValidAttribute("type"),
	    new JspUtil.ValidAttribute("beanName") };

	private static final JspUtil.ValidAttribute[] plugInAttrs = {
	    new JspUtil.ValidAttribute("type",true),
	    new JspUtil.ValidAttribute("code", true),
	    new JspUtil.ValidAttribute("codebase"),
	    new JspUtil.ValidAttribute("align"),
	    new JspUtil.ValidAttribute("archive"),
	    new JspUtil.ValidAttribute("height"),
	    new JspUtil.ValidAttribute("hspace"),
	    new JspUtil.ValidAttribute("jreversion"),
	    new JspUtil.ValidAttribute("name"),
	    new JspUtil.ValidAttribute("vspace"),
	    new JspUtil.ValidAttribute("width"),
	    new JspUtil.ValidAttribute("nspluginurl"),
	    new JspUtil.ValidAttribute("iepluginurl") };

	/*
	 * Constructor
	 */
	ValidateVisitor(JspC compiler) 
        {
	    this.pageInfo = compiler.getPageInfo();
	    //this.err = compiler.getErrorHandler();
	}

	public void visit(Node.JspRoot n)  
            throws JasperException, JspToJavaException 
        {
	    JspUtil.checkAttributes("Jsp:root", n.getAttributes(),jspRootAttrs, n.getStart()); 
	    visitBody(n);
	}

	public void visit(Node.IncludeDirective n)  
            throws JasperException, JspToJavaException  
        {
	    JspUtil.checkAttributes("Include directive", n.getAttributes(),includeDirectiveAttrs, n.getStart()); 
	    visitBody(n);
	}

	public void visit(Node.TaglibDirective n)  
            throws JasperException, JspToJavaException  
        {
	    JspUtil.checkAttributes("Taglib directive", n.getAttributes(), taglibDirectiveAttrs, n.getStart()); 
	}

	public void visit(Node.ParamAction n) 
            throws JasperException, JspToJavaException 
        {
	    JspUtil.checkAttributes("Param action", n.getAttributes(),paramActionAttrs, n.getStart());
	    n.setValue(getJspAttribute("value", n.getAttributeValue("value"),
				       n.isXmlSyntax()));
	}

	public void visit(Node.IncludeAction n) 
            throws JasperException, JspToJavaException  
        {
	    JspUtil.checkAttributes("Include action", n.getAttributes(), includeActionAttrs, n.getStart());
	    n.setPage(getJspAttribute("page", n.getAttributeValue("page"),
				      n.isXmlSyntax()));
	    visitBody(n);
        };

	public void visit(Node.ForwardAction n) 
            throws JasperException, JspToJavaException  
        {
            JspUtil.checkAttributes("Forward", n.getAttributes(), forwardActionAttrs, n.getStart());
	    n.setPage(getJspAttribute("page", n.getAttributeValue("page"),
				      n.isXmlSyntax()));
	    visitBody(n);
	}

	public void visit(Node.GetProperty n) 
            throws JasperException, JspToJavaException  
        {
	    JspUtil.checkAttributes("GetProperty", n.getAttributes(), getPropertyAttrs, n.getStart());
	}

	public void visit(Node.SetProperty n) 
            throws JasperException, JspToJavaException 
        {
	    JspUtil.checkAttributes("SetProperty", n.getAttributes(),
				    setPropertyAttrs, n.getStart());
	    String name = n.getAttributeValue("name");
	    String property = n.getAttributeValue("property");
	    String param = n.getAttributeValue("param");
	    String value = n.getAttributeValue("value");

	    if ("*".equals(property)) { 
		if (param != null || value != null)
                    throw new JspToJavaException(n.getStart(), "Invalid SetProperty");
		    
		
	    } 
            else if (param != null && value != null) 
            {
                throw new JspToJavaException(n.getStart(), "Invalid SetProperty");
		
	    }
	    n.setValue(getJspAttribute("value", value, n.isXmlSyntax()));
	}

	public void visit(Node.UseBean n) 
            throws JasperException, JspToJavaException  
        {
            //System.out.println("vist useBean");
	    JspUtil.checkAttributes("UseBean", n.getAttributes(),
				    useBeanAttrs, n.getStart());

	    String name = n.getAttributeValue ("id");
	    String scope = n.getAttributeValue ("scope");
	    String className = n.getAttributeValue ("class");
	    String type = n.getAttributeValue ("type");
	    BeanRepository beanInfo = pageInfo.getBeanRepository();

	    if (className == null && type == null)
                throw new JspToJavaException(n.getStart(), "Missing type in UseBean");
            

	    if (beanInfo.checkVariable(name))
                throw new JspToJavaException(n.getStart(), "Duplicate UseBean");
		

	    if ("session".equals(scope) && !pageInfo.isSession())
                throw new JspToJavaException(n.getStart(), "UseBean : no session");
		

	    Node.JspAttribute jattr
		= getJspAttribute("beanName", n.getAttributeValue("beanName"),
				  n.isXmlSyntax());
	    n.setBeanName(jattr);
	    if (className != null && jattr != null)
                throw new JspToJavaException(n.getStart(), "UseBean error, not both");
		

	    if (className == null)
		className = type;

            //System.out.println("Bean : "+name+"    Class="+className);
	    if (scope == null || scope.equals("page")) 
            {
		beanInfo.addPageBean(name, className);
	    } 
            else if (scope.equals("request")) 
            {
		beanInfo.addRequestBean(name, className);
	    } 
            else if (scope.equals("session")) 
            {
		beanInfo.addSessionBean(name,className);
	    } 
            else if (scope.equals("application")) 
            {
		beanInfo.addApplicationBean(name,className);
	    } 
            else 
            {
                throw new JspToJavaException(n.getStart(), "UseBean error, bad scope");
                
            }
	    visitBody(n);
	}

	public void visit(Node.PlugIn n) 
            throws JasperException, JspToJavaException  
        {
	    JspUtil.checkAttributes("Plugin", n.getAttributes(),
				    plugInAttrs, n.getStart());

	    String type = n.getAttributeValue("type");
	    if (type == null)
                throw new JspToJavaException(n.getStart(), "Plugin error, no type");
		
	    if (!type.equals("bean") && !type.equals("applet"))
                throw new JspToJavaException(n.getStart(), "Plugin error, bad type");
		
	    if (n.getAttributeValue("code") == null)
                throw new JspToJavaException(n.getStart(), "Plugin error, no code");
		

	    n.setHeight(getJspAttribute("height", n.getAttributeValue("height"),
				      n.isXmlSyntax()));
	    n.setWidth(getJspAttribute("width", n.getAttributeValue("width"),
				      n.isXmlSyntax()));
	    visitBody(n);
	}

	public void visit(Node.CustomTag n) 
            throws JasperException, JspToJavaException  
        {
            //System.out.println("Validator.ValidateVisitor visit(CustomTag)");
            
            
	    TagLibraryInfo tagLibInfo = (TagLibraryInfo) pageInfo.getTagLibraries().get(n.getPrefix());
            
            if (tagLibInfo==null)
            {
               //System.out.println("Validator.ValidateVisitor TagLibInfo is null!"); 
            }
            else
            {
               //System.out.println("Validator.ValidateVisitor TagLibInfo is "+tagLibInfo.getPrefixString());  
            }
            
            
	    TagInfo tagInfo = tagLibInfo.getTag(n.getShortName());
	    if (tagInfo == null) 
            {   
                throw new JspToJavaException(n.getStart(), "TagInfo error, missing tag info");	
	    }
            

	    /*
	     * Make sure all required attributes are present
	     */
	    TagAttributeInfo[] tldAttrs = tagInfo.getAttributes();
	    Attributes attrs = n.getAttributes();
	    for (int i=0; i<tldAttrs.length; i++) 
            {
		if (tldAttrs[i].isRequired() && attrs.getValue(tldAttrs[i].getName()) == null) 
                {
                    throw new JspToJavaException(n.getStart(), "TagInfo error, missing attribute name="+tldAttrs[i].getName()+" shortname="+n.getShortName());	
                    
		}
	    }

	    /*
	     * Make sure there are no invalid attributes
	     */
	    Hashtable tagDataAttrs = new Hashtable(attrs.getLength());
	    Node.JspAttribute[] jspAttrs
		= new Node.JspAttribute[attrs.getLength()];
	    for (int i=0; i<attrs.getLength(); i++) {
		boolean found = false;
		for (int j=0; j<tldAttrs.length; j++) {
		    if (attrs.getQName(i).equals(tldAttrs[j].getName())) {
			if (tldAttrs[j].canBeRequestTime()) {
			    jspAttrs[i]
				= getJspAttribute(attrs.getQName(i),
						  attrs.getValue(i),
						  n.isXmlSyntax());
			} else {
			    jspAttrs[i]
				= new Node.JspAttribute(attrs.getQName(i),
							attrs.getValue(i),
							false);
			}
			if (jspAttrs[i].isExpression()) {
			    tagDataAttrs.put(attrs.getQName(i),
					     TagData.REQUEST_TIME_VALUE);
			} else {
			    tagDataAttrs.put(attrs.getQName(i),
					     attrs.getValue(i));
			}
			found = true;
			break;
		    }
		}
		if (!found) 
                {
                    throw new JspToJavaException(n.getStart(), "TagInfo error, bad attribute, name="+attrs.getQName(i));	
		}
	    }

	    TagData tagData = new TagData(tagDataAttrs);
	    n.setTagData(tagData);
	    n.setJspAttributes(jspAttrs);

	    visitBody(n);
	}

	/**
	 * Preprocess attributes that can be expressions.  Expression
	 * delimiters are stripped.
	 */
	private Node.JspAttribute getJspAttribute(String name,
						  String value,
						  boolean isXml) 
        {
	    // XXX Is it an error to see "%=foo%" in non-Xml page?
	    // (We won't see "<%=foo%> in xml page because '<' is not a
	    // valid attribute value in xml).

	    if (value == null)
		return null;

	    if (isXml && value.startsWith("%=")) {
		return new Node.JspAttribute(name,
					     value.substring(2,
							     value.length()-1),
					     true);
	    }

	    if (!isXml && value.startsWith("<%=")) {
		return new Node.JspAttribute(name,
					     value.substring(3,
							     value.length()-2),
					     true);
	    }

	    return new Node.JspAttribute(name, value, false);
	}
    }

    /**
     * A visitor for validating TagExtraInfo classes of all tags
     */
    static class TagExtraInfoVisitor extends Node.Visitor 
    {

	private PageInfo pageInfo;
	//private JspErrorHandler err;
        

	/*
	 * Constructor
	 */
	TagExtraInfoVisitor(JspC compiler) 
        {
	    this.pageInfo = compiler.getPageInfo();
	    //this.err = compiler.getErrorHandler();
	}

	public void visit(Node.CustomTag n) 
            throws JasperException, JspToJavaException   
        {
	    TagLibraryInfo tagLibInfo = (TagLibraryInfo)
		pageInfo.getTagLibraries().get(n.getPrefix());
	    TagInfo tagInfo = tagLibInfo.getTag(n.getShortName());
	    if (tagInfo == null) 
            {
                throw new JspToJavaException(n.getStart(), "Missing TagInfo, name="+n.getName());
	    }

	    if (!tagInfo.isValid(n.getTagData())) 
            {
                throw new JspToJavaException(n.getStart(), "TagInfo, missing attributes");
	    }

	    visitBody(n);
	}
    }

    public static void validate(JspC compiler, Node.Nodes page) 
        throws JasperException, JspToJavaException  
    {

	/*
	 * Visit the page directives first, as they are global to the page
	 * and are position independent.
	 */
	page.visit(new PageDirectiveVisitor(compiler));

	// Determine the default output content type, per errata_a
	// http://jcp.org/aboutJava/communityprocess/maintenance/jsr053/errata_1_2_a_20020321.html
        
	PageInfo pageInfo = compiler.getPageInfo();
	String contentType = pageInfo.getContentType();
	if (contentType == null || contentType.indexOf("charset=") < 0) {
	    boolean isXml = page.getRoot().isXmlSyntax();
	    String defaultType;
	    if (contentType == null) {
		defaultType = isXml? "text/xml": "text/html";
	    } else {
		defaultType = contentType;
	    }
	    String charset = pageInfo.getPageEncoding();
	    if (charset == null)
		charset = isXml? "UTF-8": "ISO-8859-1";
	    pageInfo.setContentType(defaultType + ";charset=" + charset);
	}

	/*
	 * Validate all other nodes.
	 * This validation step includes checking a custom tag's mandatory and
	 * optional attributes against information in the TLD (first validation
	 * step for custom tags according to JSP.10.5).
	 */
	page.visit(new ValidateVisitor(compiler));

	/*
	 * Invoke TagLibraryValidator classes of all imported tags
	 * (second validation step for custom tags according to JSP.10.5).
	 */
	validateXmlView(new PageDataImpl(page), compiler);

	/*
	 * Invoke TagExtraInfo method isValid() for all imported tags 
	 * (third validation step for custom tags according to JSP.10.5).
	 */
	page.visit(new TagExtraInfoVisitor(compiler));

    }


    //*********************************************************************
    // Private (utility) methods

    /**
     * Validate XML view against the TagLibraryValidator classes of all
     * imported tag libraries.
     */
    private static void validateXmlView(PageData xmlView, JspC compiler)
	        throws JasperException, JspToJavaException  
    {

	StringBuffer errMsg = null;
	//JspErrorHandler errDisp = compiler.getErrorHandler();

        //TODO TODO OBS Taglibrary
        Enumeration enum = compiler.getPageInfo().getTagLibraries().elements();
        while (enum.hasMoreElements()) 
        {
            TagLibraryInfo tli = (TagLibraryInfo) enum.nextElement();
            
            //ValidationMessage[] errors = ((TagLibraryInfoImpl) tli).validate(xmlView);
	    ValidationMessage[] errors = null; 
            if ((errors != null) && (errors.length != 0)) 
            {
                if (errMsg == null) {
		    errMsg = new StringBuffer();
		}
                errMsg.append("<h3>");
                errMsg.append("jsp.error.tlv.invalid.page"+tli.getShortName());
                errMsg.append("</h3>");
                for (int i=0; i<errors.length; i++) {
                    errMsg.append("<p>");
                    errMsg.append(errors[i].getId());
                    errMsg.append(": ");
                    errMsg.append(errors[i].getMessage());
                    errMsg.append("</p>");
                }
            }
        }

	if (errMsg != null) 
        {
            throw new JspToJavaException(errMsg.toString());
            //errDisp.jspError(errMsg.toString());
	}
    }
}

