/* 
 * Copyright (c) 2001 by Matt Welsh and The Regents of the University of 
 * California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Authors: Eric Wagner <eric@xcf.berkeley.edu> 
 *          Matt Welsh <mdw@cs.berkeley.edu> 
 */


package com.rimfaxe.webserver.seda;

import seda.apps.Haboob.hdapi.*;

import seda.apps.Haboob.*;
import seda.apps.Haboob.http.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.lib.http.*;
import seda.sandStorm.lib.aDisk.*;
import seda.util.*;
import java.io.*;
import java.util.*;

public class DynamicHttp implements EventHandlerIF, HaboobConst 
{

  private static final boolean DEBUG = false;

  // Whether to close the HTTP connection after each request
  private static final boolean CLOSE_CONNECTION = false;

  // Whether each URL should be handled by its own stage
  private static final boolean SEPARATE_STAGES = true;

  private SinkIF mysink;
  private static ConfigDataIF config;
  private static SinkIF mainsink;
  private static Hashtable handlerCache, stageCache;
  private String myurl;

  public DynamicHttp() 
  {
    myurl = null;
  }

  private DynamicHttp(String myurl) 
  {
    this.myurl = myurl;
  }

  public void init(ConfigDataIF config) 
    throws Exception 
  {
    this.config = config;
    mysink = config.getStage().getSink();

    if (myurl == null) 
    {
      mainsink = mysink;
      handlerCache = new Hashtable();
      stageCache = new Hashtable();
    } 
    else 
    {
      //System.out.println("DynamicHttp handlerStage ["+myurl+"]: Started");
    }
  }

  public void destroy() {
  }

  /**
   * Handle the given request, passing it to the appropriate handler
   * (and stage, if necessary). Returns true if the request can be 
   * processed, or false if the request was not for a dynamic URL.
   */
  public static boolean handleRequest(httpRequest req) 
    throws Exception 
  {
    HaboobStats.numRequests++;
    String url = req.getURL();
    
    //System.out.println("DynamicHttp: handle "+url);
    
   
    boolean isStatic = false;
    
    String classname = "com.rimfaxe.webserver.seda.SedaHandler";
    try 
    {
      handlerPool pool = (handlerPool) handlerCache.get(classname);
      
     
      httpRequestHandlerIF handler = null;
      
      if (pool == null) 
      {
	handler = new com.rimfaxe.webserver.seda.SedaHandler();
      }
      else
      {
        handler = pool.getHandler();   
      }
      
          
      isStatic = handler.isCacheable(req);
      
    } 
    catch (Exception e) 
    {
      System.out.println(""+e);  
      isStatic = false;
    }
    
    if (isStatic) return false;
    

    if (SEPARATE_STAGES) 
    {
      SinkIF thesink;
      synchronized (stageCache) 
      {
	thesink = (SinkIF)stageCache.get(url);
	if (thesink == null) thesink = makeStage(url);
      }
      thesink.enqueue(req);
    } 
    else 
    {
      mainsink.enqueue(req);
    }
    return true;
  }

  public void handleEvent(QueueElementIF item) 
  {
    if (DEBUG) 
    {
      if (myurl == null) 
        System.err.println("DynamicHttp: GOT QEL: " + item);
      else 
        System.err.println("DynamicHttp ["+myurl+"]: GOT QEL: " + item);
    }

    if (item instanceof httpRequest) 
    {
      httpRequest req = (httpRequest) item;
      try 
      {
	doRequest(req);
      } 
      catch (Exception e) 
      {
        com.rimfaxe.webserver.ErrorLog.log("DynamicHttp", 170, "Internal error", 
           ""+req+"\n"+e);
        
	HttpSend.sendResponse(
	    new httpResponder(
	      new httpInternalServerErrorResponse(req, e, "DynamicHttp.handleEvent()") , req, true)); 
      }

    } 
    else 
    {
      System.err.println( "DynamicHttp: Don't know what to do with " + item );
    }
  }

  public void handleEvents(QueueElementIF items[]) 
  {
    for (int i = 0; i < items.length; i++) 
    {
      handleEvent(items[i]);
    }
  }

  private void doRequest(httpRequest req) 
  {
    String url;
    String classname;

    
    
    url = req.getURL();
    

    classname = "com.rimfaxe.webserver.seda.SedaHandler";
    
    StringBuffer error_log = new StringBuffer();
    
    try 
    {
      
       
       
      
      
      handlerPool pool;
      
      synchronized (this) 
      {
                  
	pool = (handlerPool) handlerCache.get(classname);
	if (pool == null) 
        {
           
	  try 
          {
             
	    pool = new handlerPool(classname);
            if (DEBUG)
            {
	      System.err.println("DynamicHttp: Loaded class " + classname + " for url " + url);
            }
	  } 
          catch (ClassNotFoundException cnfe) 
          {
            com.rimfaxe.webserver.ErrorLog.log("DynamicHttp", 226, "Internal error", 
           ""+req+"\n"+cnfe);
            
	    HttpSend.sendResponse(
		new httpResponder(
		  new httpInternalServerErrorResponse(req, cnfe, "DynamicHttp.doRequest() - pool class"), req, true)); 
	    return;
	  }
          
	  handlerCache.put(classname, pool);
	}
      }

      
      
      
      httpRequestHandlerIF handler = pool.getHandler();
      
      httpResponse resp;
      
      
      
      resp = handler.handleRequest(req,error_log);
      //resp = handler.handleRequest(req);
       
      resp.setServerPort( req.getServerPort() );     
     
      boolean shouldClose = CLOSE_CONNECTION;
      if (req.getHttpVer()!=req.HTTPVER_11) shouldClose = true;
      
      
      httpResponder respd = new httpResponder(resp, req, shouldClose);
      HttpSend.sendResponse(respd);
      pool.doneWithHandler(handler);
      
      
      //System.out.println("DynamicHttp : "+error_log+"\n"+req+"\n"); 
       
    } 
    catch (Exception e) 
    {
      com.rimfaxe.webserver.ErrorLog.log("DynamicHttp", 170, "Internal error", 
           ""+req+"\n"+e+"\n"+error_log+"\n");
     
  
      HttpSend.sendResponse( new httpResponder(  new httpInternalServerErrorResponse(req, e, "DynamicHttp.doRequest()"), req, true)); 
      return;
    }
  }

  class handlerPool 
  {
    Class theclass;
    Vector pool;

    handlerPool(String classname) 
        throws ClassNotFoundException 
    {
      this.theclass = Class.forName(classname);
      this.pool = new Vector();
    }

    httpRequestHandlerIF getHandler() 
        throws InstantiationException, IllegalAccessException 
    {
      httpRequestHandlerIF handler;
      synchronized (pool) 
      {
	if (pool.size() == 0) 
        {
	  handler = (httpRequestHandlerIF) theclass.newInstance();
	} 
        else 
        {
	  handler = (httpRequestHandlerIF) pool.remove(pool.size() - 1);
	}
      }
      return handler;
    }

    void doneWithHandler(httpRequestHandlerIF handler) 
    {
      synchronized (pool) 
      {
	pool.addElement(handler);
      }
    }
  }

  private static SinkIF makeStage(String url) 
      throws Exception 
  {
    StageIF thestage;
    thestage = config.getManager().createStage("DynamicHttp ["+url+"]", new DynamicHttp(url), null);
    stageCache.put(url, thestage.getSink());
    return thestage.getSink();
  }

  

  
}
