/*
 * RimfaxeHttpServletRequest.java
 *
 *
 * Copyright (c) 2003 Rimfaxe ApS  (www.rimfaxe.com).  
 * All rights reserved.
 *
 * This package is written by Lars Andersen <lars@rimfaxe.com> 
 * and licensed by Rimfaxe ApS.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by Rimfaxe ApS
          (www.rimfaxe.com)"
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Rimfaxe", "Rimfaxe Software", "Lars Andersen" and 
 *    "Rimfaxe WebServer" must not be used to endorse or promote products 
 *    derived from this software without prior written permission. For written
 *    permission, please contact info@rimfaxe.com
 *
 * 5. Products derived from this software may not be called "Rimfaxe"
 *    nor may "Rimfaxe" appear in their names without prior written
 *    permission of the Rimfaxe ApS.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.rimfaxe.webserver.servletapi;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.Principal;
import javax.servlet.*;
import javax.servlet.http.*;

import seda.sandStorm.lib.http.httpRequest;


import com.rimfaxe.webserver.servletapi.session.SessionStore;
import com.rimfaxe.webserver.*;



public class RimfaxeHttpServletRequest implements HttpServletRequest 
{

    WebContext webcontext;   
    SessionStore sessionstore;
    
    /** getHttpVer() code corresponding to HTTP/0.9. */
    public static final int HTTPVER_09 = 0;
    /** getHttpVer() code corresponding to HTTP/1.0. */
    public static final int HTTPVER_10 = 1;
    /** getHttpVer() code corresponding to HTTP/1.1. */
    public static final int HTTPVER_11 = 2;
  
    
    /**
     * The initial state of the request InputStream 
     */
    private final static int STREAM_STATE_INITIAL = 0;

    /**
     * One reader has been created and is probably used.
     */
    private final static int STREAM_READER_USED = 1;

    /**
     * The input stream is used
     */
    private final static int INPUT_STREAM_USED = 2;

    /**
     * The inputstream state
     */
    private int stream_state = STREAM_STATE_INITIAL;

    public final static String STATE_PARAMETERS = "com.rimfaxe.servlet.stateParam";

    /** 
     * The initial request.
     */
    private httpRequest request = null;
    
    /**
     * The attached servlet.
     */
    private Servlet servlet = null;
    
    
    private Hashtable postparameters = new Hashtable();
    private Hashtable attributes = new Hashtable();

    protected RimfaxeHttpServletResponse response = null;
    protected RimfaxeHttpSession httpSession = null;

    protected String requestedSessionID = null;

    
    /**
   * Return the code corresponding to the HTTP version. Each code has
   * one of the HTTPVER_* values from this class.
   */
    public int getHttpVer() 
    {
      return request.getHttpVer();
    }
  

    /**
     * Return the Charset parameter of content type
     * @return A String instance
     */
    public String getCharacterEncoding() 
    {
	return "iso-8859-1";
    }

    /**
     * ServletRequest implementation - Get the length of request data.
     * @return An int, or <strong>-1</strong>.
     */
    public int getContentLength() 
    {
	return -1;
    }

    /**
     * ServletRequest implementation - Get the type of the request's body.
     * @return A String encoded mime type, or <strong>null</strong>.
     */ 
    public String getContentType() 
    {
	return request.getHeader("Content-Type");
    }

    /**
     * ServletRequest implementation - Get the protocol of that request.
     * @return A String encoded version of the protocol.
     */
    public String getProtocol() 
    {
        
	return "HTTP";
    }

    

   /**
    * ServletRequest implementation - Get the name of queried server.
    * @return Name of server, as a String.
    */
   public String getServerName() 
   {
       
       String host = request.getHeader("Host");
       if (host != null) 
       {
	   int idx = host.lastIndexOf(':');
	   if (idx != -1) 
           {
	       return host.substring(0, host.lastIndexOf(':'));
	   } 
           else 
           {
	       return host;
	   }
       } 
       else 
       {
	   return "localhost";
       }
   }

   /**
    * ServletRequest implementation - Get the port of queried server.
    * @return A port number (int).
    */
   public int getServerPort() 
   {
       return request.getServerPort();  
   }

   /**
    * ServletRequest implementation - Get the IP address of requests's sender.
    * @return Numeric IP address, as a String.
    */

    public String getRemoteAddr() 
    {
        //return "unknown";
	return request.getConnection().getConnection().getAddress().getHostAddress();
    }

    /**
     * ServletRequest implementation - FQDN of request's sender.
     * @return Name of client's machine (FQDN).
     */
    public String getRemoteHost() 
    {
         
        
        //return "unknown";
	return request.getConnection().getConnection().getAddress().getHostName();
    }

    
    public String getRealPath(String name) 
    {
       
       String ctxt_root = ""; 
       if (request.getURL().startsWith("/"))
       {
         ctxt_root = webcontext.getUrlpath().substring(0,webcontext.getUrlpath().length()-1);
       }
       else
       {
         ctxt_root = webcontext.getUrlpath();  
       }
        return request.getURL();
    }

    public String prependRealPath(String name) 
    {
      String ctxt_root = webcontext.getUrlpath().substring(0,webcontext.getUrlpath().length()-1);  
      return ctxt_root + name;
    }
    
    
    
    public ServletInputStream getInputStream()
	throws IOException
    {
	if (stream_state == STREAM_READER_USED)
	    throw new IllegalStateException("Reader used");
	stream_state = INPUT_STREAM_USED;
	return null;
    }

    

    /**
     * ServletRequest implementation - Get a parameter value.
     * @return The String encoded value for the parameter.
     */

    public String getParameter(String name) 
    {
      //System.out.println("Get Parameter  -  "+name+"="+request.getQuery(name));  
      return request.getQuery(name);
    }

    /**
     * ServletRequest implementation - Get the parameters value.
     * @return The String array encoded value for the parameter.
     */
    public String[] getParameterValues(String parameter) 
    {
      //System.out.println("Get ParameterValues");  
      return request.getQuerySet(parameter);
    }

    /**
     * ServletRequest implementation - List available parameters.
     * @return An enumeration of parameter names.
     */

    public Enumeration getParameterNames() 
    {
        //System.out.println("Get ParameterNames"); 
        Enumeration enum = request.getQueryKeys();
        
        if (enum==null) 
        {
          //System.out.println("Get ParameterNames - enum is null");
          Vector vec = new Vector();
          enum = vec.elements();
        }
        
        
	return enum;
    }

    
    public Object getAttribute(String name) 
    {
	return attributes.get(name);
    }

    public void setAttribute(String name, Object object) 
    {
	attributes.put(name, object);
    }

    /**
     * Removes an attribute from this request.  This method is not
     * generally needed as attributes only persist as long as the request
     * is being handled.
     *
     * <p>Attribute names should follow the same conventions as
     * package names. Names beginning with <code>java.*</code>,
     * <code>javax.*</code>, and <code>com.sun.*</code>, are
     * reserved for use by Sun Microsystems.
     *
     * @param name a <code>String</code> specifying 
     * the name of the attribute to remove
     */
    public void removeAttribute(String name) 
    {
	attributes.remove(name);
    }

    public Enumeration getAttributeNames() 
    {
	return attributes.keys();
    }

    /**
     * Returns the preferred <code>Locale</code> that the client will 
     * accept content in, based on the Accept-Language header.
     * If the client request doesn't provide an Accept-Language header,
     * this method returns the default locale for the server.
     *
     * @return the preferred <code>Locale</code> for the client
     */
    public Locale getLocale() 
    {
	return (Locale) getLocales().nextElement();
    }

    /**
     * Returns an <code>Enumeration</code> of <code>Locale</code> objects
     * indicating, in decreasing order starting with the preferred locale, the
     * locales that are acceptable to the client based on the Accept-Language
     * header.
     * If the client request doesn't provide an Accept-Language header,
     * this method returns an <code>Enumeration</code> containing one 
     * <code>Locale</code>, the default locale for the server.
     *
     * @return an <code>Enumeration</code> of preferred 
     * <code>Locale</code> objects for the client
     */
    public Enumeration getLocales() 
    {
        Vector locales = new Vector(); 
        
        String accept_lang = getHeader("Accept-Language");
        
        if (accept_lang!=null)
        {
          StringTokenizer tkz = new StringTokenizer(accept_lang, " ,",false);
          while (tkz.hasMoreTokens())
          {
            String lang = tkz.nextToken();
            RimfaxeLocale ql = new RimfaxeLocale(lang.trim());
            locales.addElement( ql.getLocale() );
          }
        }
        else
        {
          locales.addElement( Locale.getDefault() );   
        }
        
	return locales.elements();
    }

    /**
     * Returns a boolean indicating whether this request was made using a
     * secure channel, such as HTTPS.
     *
     * @return a boolean indicating if the request was made using a
     * secure channel
     */
    public boolean isSecure() 
    {
	// only https secure?
	return false;
    }

    /**
     * HttpServletRequest implementation - Get the request's method.
     * @return A String instance.
     */
    public  String getMethod() 
    {
        //System.out.println("Get Method"); 
	return request.getMethod();
    }

    /**
     * HttpServletRequest implementation - Get the request's path info.
     * @return A String instance or <strong>null</strong>.
     */
    public  String getPathInfo() 
    {
	return null;
    }

    /**
     * HttpServletRequest implementation - Get the request's path translated.
     * @return A String instance or <strong>null</strong>.
     */
    public  String getPathTranslated() 
    {
	String pathinfo = getPathInfo();
	if ( pathinfo != null )
	    return getRealPath(pathinfo);
	return null;
    }

    /**
     * Returns the portion of the request URI that indicates the context
     * of the request.  The context path always comes first in a request
     * URI.  The path starts with a "/" character but does not end with a "/"
     * character.  For servlets in the default (root) context, this method
     * returns "".
     * @return a <code>String</code> specifying the portion of the request 
     * URI that indicates the context of the request
     */
    public String getContextPath() 
    {
        String cpath = webcontext.getUrlpath();
        if (cpath.endsWith("/"))
            return cpath.substring(0,cpath.length()-1);
	return cpath;
    }

    public boolean hasQueryString() 
    {
        if (request.getQueryString()!=null) return true;
        else return false;
    }

    /**
     * HttpServletRequest implementation - Get the request's query string.
     * @return A String instance or <strong>null</strong>.
     */
    public String getQueryString() 
    {
        
	return request.getQueryString();
    }

    /**
     * HttpServletRequest implementation - Get the request's user (if any).
     * @return A String instance or <strong>null</strong>.
     */
    public String getRemoteUser() 
    {
        
        return null;
    }

    /**
     * Returns a boolean indicating whether the authenticated user is included
     * in the specified logical "role".  Roles and role membership can be
     * defined using deployment descriptors.  If the user has not been
     * authenticated, the method returns <code>false</code>.
     *
     * @param role a <code>String</code> specifying the name of the role
     * @return a <code>boolean</code> indicating whether the user making this
     * request belongs to a given role; <code>false</code> if the user has not
     * been authenticated
     */

    public boolean isUserInRole(String role) 
    {
	throw new RuntimeException("Not Yet Implemented");
    }

    /**
     * Returns a <code>java.security.Principal</code> object containing
     * the name of the current authenticated user. If the user has not been
     * authenticated, the method returns <code>null</code>.
     *
     * @return a <code>java.security.Principal</code> containing
     * the name of the user making this request; <code>null</code> if the 
     * user has not been authenticated
     */
    public Principal getUserPrincipal() 
    {
        return null;
    }

    /**
     * HttpServletRequest implementation - Get the request's auth method.
     * @return A String instance or <strong>null</strong>.
     */

    public String getAuthType() 
    {
        return null;
    }

    /**
     * HttpServletRequest implementation - Get a request header as a String.
     * @return A String instance or <strong>null</strong>.
     */
    public String getHeader(String name) 
    {
	return request.getHeader(name);
    }

    /**
     * Returns all the values of the specified request header
     * as an <code>Enumeration</code> of <code>String</code> objects.
     *
     *
     * @param name a <code>String</code> specifying the header name
     * @return a <code>Enumeration</code> containing the values of the 
     * requested header, or <code>null</code> if the request does not
     * have any headers of that name
     */			
    public Enumeration getHeaders(String name) 
    {
	String value = getHeader(name);
        
        if (value==null) return null;
        
        Vector vec = new Vector();
        vec.addElement(value);
	
	return vec.elements();
    }

    /**
     * HttpServletRequest implementation - Get a request header as an int.
     * @return An int, or <strong>-1</strong>.
     */
    public int getIntHeader(String name) 
    {
	String v = getHeader(name);
	if ( v != null ) 
        {
            try 
            {
                Integer ival = new Integer(v);
	        return ival.intValue();
            } 
            catch (NumberFormatException e) 
            {
                System.out.println("Error parsing Header ["+name+": "+v+"] to integer, return -1 instead");   
            }
            
	}
	return -1;
    }

    /**
     * HttpServletRequest implementation - Get a request header as an date.
     * @return An long (as a number of milliseconds), or <strong>-1</strong>.
     */
    public long getDateHeader(String name) 
    {
	return (long) -1;
    }

    /**
     * HttpServletRequest implementation - Get a all header names.
     * @return An enumeration.
     */
    public Enumeration getHeaderNames() 
    {
	return request.enumerateHeaders();
    }

    /**
     * Gets, from the first line of the HTTP request, 
     * the part of this request's URI that is to the left of any query string.
     */
    public String getRequestURI() 
    {
        
	String uri = request.getURL();
	
	return uri;
    }

    /**
     * Returns a {@link RequestDispatcher} object that acts as a wrapper for
     * the resource located at the given path.  
     * A <code>RequestDispatcher</code> object can be used to forward
     * a request to the resource or to include the resource in a response.
     * The resource can be dynamic or static.
     *
     * <p>The pathname specified may be relative, although it cannot extend
     * outside the current servlet context.  If the path begins with 
     * a "/" it is interpreted as relative to the current context root.  
     * This method returns <code>null</code> if the servlet container
     * cannot return a <code>RequestDispatcher</code>.
     *
     * <p>The difference between this method and {@link
     * ServletContext#getRequestDispatcher} is that this method can take a
     * relative path.
     *
     * @param path a <code>String</code> specifying the pathname
     * to the resource
     * @return a <code>RequestDispatcher</code> object that acts as a 
     * wrapper for the resource at the specified path
     * @see RequestDispatcher
     * @see ServletContext#getRequestDispatcher
     */
    public RequestDispatcher getRequestDispatcher(String path) 
    {
	if (path == null) 
        {
	    throw new IllegalArgumentException("null");
	}
        
	//return null;
	return RimfaxeRequestDispatcher.getRequestDispatcher(path, webcontext.getVirtualHost(), webcontext);
    }

    /**
     * Gets the part of this request's URI that refers to the servlet 
     * being invoked. Analogous to the CGI variable SCRIPT_NAME. 
     */
    public String getServletPath() 
    {
      return request.getURL();
    }

    /**
     * @return the scheme of the URL used in this request, for example "http",
     * "https", or "ftp". Different schemes have different rules
     * for constructing URLs, as noted in RFC 1738. The URL used to create 
     * a request may be reconstructed using this scheme, the server name 
     * and port, and additional information such as URIs.
     */
    public String getScheme() 
    {
	return "http";
    }

    /**
     * Gets the array of cookies found in this request.
     * @return the array of cookies found in this request or
     * <strong>null</strong> if there is no cookie.
     */
    public Cookie[] getCookies() 
    {
        Vector vec = new Vector();
        String cookie_str = getHeader("Cookie");
        
        if (cookie_str==null) return null;
        
        StringTokenizer tkz = new StringTokenizer(cookie_str," ;",false);
        while (tkz.hasMoreTokens())
        {
           String ck = tkz.nextToken();
           StringTokenizer tkz2 = new StringTokenizer(ck,"=",false);
           String name = tkz2.nextToken();
           String value = tkz2.nextToken();
           vec.addElement( new Cookie(name,value) );
        }
        
        Cookie[] clist = new Cookie[vec.size()];
        for (int i=0; i<vec.size(); i++)
        {
          clist[i] = (Cookie) vec.elementAt(i);  
        }
	
	return clist;
    }


    protected String getRequestedSessionIdFromCookie() 
    {
	Cookie[] clist = getCookies();
        if (clist==null) return null;
        
        for (int i=0; i<clist.length; i++)
        {
          if (clist[i].getName().equalsIgnoreCase(getCookieName())) 
              return clist[i].getValue();
        }
        
	return null;
    }

    protected String getRequestedSessionIdFromURL() 
    {
	return getParameter(getCookieName());
    }

    /**
     * Gets the session id specified with this request. This may differ 
     * from the actual session id. For example, if the request specified an
     * id for an invalid session, then this will get a new session with a 
     * new id. 
     * @return the session id specified by this request, or null if the 
     * request did not specify a session id.
     */
    public String getRequestedSessionId() 
    {
	if (requestedSessionID == null) 
        {
	    requestedSessionID = getRequestedSessionIdFromCookie();
	    if (requestedSessionID == null)
		requestedSessionID = getRequestedSessionIdFromURL();
	}
	return requestedSessionID;
    }

    
    /**
     * Gets the current valid session associated with this request, if create
     * is false or, if necessary, creates a new session for the request, if 
     * create is true. 
     * @return the session associated with this request or null if create 
     * was false and no valid session is associated with this request. 
     */
    public HttpSession getSession(boolean create) 
    {
	if (httpSession == null) 
        {
	    httpSession = (RimfaxeHttpSession) getSession(getRequestedSessionId());
	    if (httpSession != null) httpSession.setNoMoreNew();
	}
	if (httpSession == null & create) 
        {
	    httpSession = new RimfaxeHttpSession(createCookie(), webcontext);
	    response.addCookie(httpSession.getCookie());
            sessionstore.putSessionObject(httpSession.getCookie().getValue(), (RimfaxeHttpSession) httpSession);
	} 
        else if (httpSession != null) 
        {
	    httpSession.setLastAccessedTime();
	    if (! httpSession.isValid()) 
            {
		httpSession = new RimfaxeHttpSession(createCookie(), webcontext);
		response.addCookie(httpSession.getCookie());
                sessionstore.putSessionObject(httpSession.getCookie().getValue(), (RimfaxeHttpSession) httpSession);
	    }
	}
	return httpSession;
    }

    /**
     * Gets the current valid session associated with this request.
     * @return the session associated with this request.
     */
    public HttpSession getSession() 
    {
	return getSession(true);
    }

    protected String getCookieName() 
    {
      return "RWS"; 
    }

    protected Cookie createCookie() 
    {
	
	String name     = getCookieName();
        
	String path     = webcontext.getUrlpath(); //"/";
	String domain   = null;
	String comment  = null;
	int maxage      = 86400;
	boolean secure  = false;

	Cookie cookie = new Cookie(name, sessionstore.getNewCookieValue()); 
	cookie.setPath(path);
	cookie.setMaxAge(maxage);
	if ((comment != null) && (comment.length() > 0))
	    cookie.setComment(comment);
	if ((domain != null) && (domain.length() > 0))
	    cookie.setDomain(domain);
	cookie.setSecure(secure);
	return cookie;
    }

    protected HttpSession getSession(String sessionId) 
    {	
       RimfaxeHttpSession rhs = sessionstore.getSessionObject(sessionId);
       
       return rhs;
    }

    /**
     * Checks whether this request is associated with a session that is valid 
     * in the current session context. If it is not valid, the requested
     * session will never be returned from the getSession method. 
     * @return true if this request is assocated with a session that is valid 
     * in the current session context. 
     */
    public boolean isRequestedSessionIdValid() 
    {
	RimfaxeHttpSession session = (RimfaxeHttpSession) getSession(getRequestedSessionId());
	if (session == null)
	    return false;
	return (session.isValid());
    }

    /**
     * Checks whether the session id specified by this request came in as 
     * a cookie. (The requested session may not be one returned by the 
     * getSession method.)
     * @return true if the session id specified by this request came 
     * in as a cookie; false otherwise 
     */
    public boolean isRequestedSessionIdFromCookie() 
    {
	return (getRequestedSessionIdFromCookie() != null);
    }

    /**
     * Checks whether the session id specified by this request came in as 
     * part of the URL. (The requested session may not be the one returned 
     * by the getSession method.)
     * @return true if the session id specified by the request for this 
     * session came in as part of the URL; false otherwise
     * @deprecated since jsdk2.1
     */
    public boolean isRequestedSessionIdFromUrl() {
	return (getRequestedSessionIdFromURL() != null);
    }

    /**
     * Checks whether the session id specified by this request came in as 
     * part of the URL. (The requested session may not be the one returned 
     * by the getSession method.)
     * @return true if the session id specified by the request for this 
     * session came in as part of the URL; false otherwise
     */
    public boolean isRequestedSessionIdFromURL() {
	return (getRequestedSessionIdFromURL() != null);
    }

    protected BufferedReader reader = null;

    /**
     * Returns a buffered reader for reading text in the request body. 
     * This translates character set encodings as appropriate.
     * @exception UnsupportedEncodingException if the character set encoding 
     * is unsupported, so the text can't be correctly decoded. 
     * @exception IllegalStateException if getInputStream has been called on 
     * this same request. 
     * @exception IOException on other I/O related errors. 
     */
    public BufferedReader getReader()
	throws IOException
    {
	return null;
    }

    /**
     * Get the wrapped Seda Request.
     * @return the request
     */
    protected httpRequest getRequest() 
    {
	return request;
    }

    public java.util.Map getParameterMap() 
    {
      return null;  
    }
    
    public StringBuffer getRequestURL() 
    {
      StringBuffer buf = new StringBuffer();  
      String host = request.getHeader("Host");
      if (host==null) host = "127.0.0.1";
      
      buf.append(this.getProtocol()+"://"+host+this.request.getURL());
      return buf;
    }
    
    public void setCharacterEncoding(String str) 
        throws java.io.UnsupportedEncodingException 
    {
    }
    
    
    RimfaxeHttpServletRequest(Servlet servlet, 
			      httpRequest request, 
			      RimfaxeHttpServletResponse response,
                              WebContext webcontext)
			     
    {
        this.webcontext     = webcontext;
	this.servlet        = servlet;
	this.request        = request;
	this.response       = response;
	this.sessionstore   = webcontext.getVirtualHost().getSessionStore();   
    }

}