/*
 * Copyright (c) 2001 by Matt Welsh and The Regents of the University of 
 * California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Matt Welsh <mdw@cs.berkeley.edu>
 * 
 */

package seda.sandStorm.lib.http;

import seda.sandStorm.api.*;
import seda.sandStorm.lib.aSocket.*;
import seda.sandStorm.core.*;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * This is an abstract class corresponding to an HTTP response.
 * Use one of the subclasses (such as httpOKResponse or httpNotFoundResponse)
 * to push responses back to the client.
 * 
 * @author Matt Welsh
 * @see httpOKResponse
 * @see httpNotFoundResponse
 */
public abstract class httpResponse implements httpConst, QueueElementIF 
{

  /** Code corresponding to '100 Continue'. */ 
  public static final int RESPONSE_CONTINUE = 100;
  /** Code corresponding to '101 Switching Protocols'. */ 
  public static final int RESPONSE_SWITCHING_PROTOCOLS = 101;
 
  
  /** Code corresponding to '200 OK'. */ 
  public static final int RESPONSE_OK = 200;
  /** Code corresponding to '201 Created'. */ 
  public static final int RESPONSE_CREATED = 201;
  /** Code corresponding to '202 ACCEPTED'. */ 
  public static final int RESPONSE_ACCEPTED = 202;
  /** Code corresponding to '203 Non-Authoritative Information'. */ 
  public static final int RESPONSE_NON_AUTHORITATIVE_INFORMATION = 203;
  /** Code corresponding to '204 No Content'. */ 
  public static final int RESPONSE_NO_CONTENT = 204;
  /** Code corresponding to '205 Reset Content'. */ 
  public static final int RESPONSE_RESET_CONTENT = 205;
  /** Code corresponding to '206 Partial Content'. */ 
  public static final int RESPONSE_PARTIAL_CONTENT = 206;
  
  
  /** Code corresponding to '300 Multiple Choices'. */ 
  public static final int RESPONSE_MULTIPLE_CHOICES = 300;
  /** Code corresponding to '301 Moved Permanently'. */ 
  public static final int RESPONSE_REDIRECT = 301;
  /** Code corresponding to '302 Found'. */ 
  public static final int RESPONSE_FOUND = 302;
  /** Code corresponding to '303 See Other'. */ 
  public static final int RESPONSE_SEE_OTHER = 303;
  /** Code corresponding to '304 Not Modified'. */ 
  public static final int RESPONSE_NOT_MODIFIED = 304;
  /** Code corresponding to '305 Use Proxy'. */ 
  public static final int RESPONSE_USE_PROXY = 305;
  /** Code corresponding to '306 (Unused)'. */ 
  public static final int RESPONSE_RESERVED_306 = 306;
  /** Code corresponding to '307 Temporary Redirect'. */ 
  public static final int RESPONSE_TEMPORARY_REDIRECT = 307;
  
  
 
  /** Code corresponding to '400 Bad Request'. */ 
  public static final int RESPONSE_BAD_REQUEST = 400;
  /** Code corresponding to '401 Unauthorized'. */ 
  public static final int RESPONSE_UNAUTHORIZED = 401; 
  /** Code corresponding to '402 Payment Required'. */ 
  public static final int RESPONSE_PAYMENT_REQUIRED = 402;
  /** Code corresponding to '403 Forbidden'. */ 
  public static final int RESPONSE_FORBIDDEN = 403; 
  /** Code corresponding to '404 Not Found'. */ 
  public static final int RESPONSE_NOT_FOUND = 404;
  /** Code corresponding to '405 Method Not Allowed'. */ 
  public static final int RESPONSE_METHOD_NOT_ALLOWED = 405; 
  /** Code corresponding to '406 Not Acceptable'. */ 
  public static final int RESPONSE_NOT_ACCEPTABLE = 406;
  /** Code corresponding to '407 Proxy Authentication Required'. */ 
  public static final int RESPONSE_PROXY_AUTHENTICATION_REQUIRED = 407; 
  /** Code corresponding to '408 Request Timeout'. */ 
  public static final int RESPONSE_REQUEST_TIMEOUT = 408;
  /** Code corresponding to '409 Conflict'. */ 
  public static final int RESPONSE_CONFLICT = 409; 
  /** Code corresponding to '410 Gone'. */ 
  public static final int RESPONSE_GONE = 410;
  /** Code corresponding to '411 Length Required'. */ 
  public static final int RESPONSE_LENGTH_REQUIRED = 411; 
  /** Code corresponding to '412 Precondition Failed'. */ 
  public static final int RESPONSE_PRECONDITION_FAILED = 412;
  /** Code corresponding to '413 Request Entity Too Large'. */ 
  public static final int RESPONSE_REQUEST_ENTITY_TOO_LARGE = 413;
  /** Code corresponding to '414 Request-URI Too Long'. */ 
  public static final int RESPONSE_REQUEST_URI_TOO_LONG = 414; 
  /** Code corresponding to '415 Unsupported Media Type'. */ 
  public static final int RESPONSE_UNSUPPORTED_MEDIA_TYPE = 415;
  /** Code corresponding to '416 Requested Range Not Satisfiable'. */ 
  public static final int RESPONSE_REQUESTED_RANGE_NOT_SATISFIABLE = 416; 
  /** Code corresponding to '417 Expectation Failed'. */ 
  public static final int RESPONSE_EXPECTATION_FAILED = 417;
  
  
  
  /** Code corresponding to '500 Internal Server Error'. */ 
  public static final int RESPONSE_INTERNAL_SERVER_ERROR = 500;
  /** Code corresponding to '501 Not Implemented'. */ 
  public static final int RESPONSE_NOT_IMPLEMENTED = 501;
  /** Code corresponding to '502 Bad Gateway'. */ 
  public static final int RESPONSE_BAD_GATEWAY = 502;
  /** Code corresponding to '503 Service Unavailable'. */ 
  public static final int RESPONSE_SERVICE_UNAVAILABLE = 503;

  /** Code corresponding to '504 Gateway Timeout'. */ 
  public static final int RESPONSE_GATEWAY_TIMEOUT = 504;
  /** Code corresponding to '505 HTTP Version Not Supported'. */ 
  public static final int RESPONSE_HTTP_VERSION_NOT_SUPPORTED = 505;
  
  
  /** The default MIME type for responses, which is "text/html". */
  public static final String DEFAULT_MIME_TYPE = "text/html";

  private static final boolean DEBUG = false;

  /** The code corresponding to the response. */
  protected int code;
  /** The default response header. */
  protected static String defaultHeader = "Server: Rimfaxe Web Server 1.03"+CRLF;

  /** The actual data of the response. */
  protected BufferElement combinedData;
  /** The header for the response. */
  protected BufferElement header;
  /** The payload for the response. */
  protected BufferElement payload;
  /** The MIME type of the response. */
  protected String contentType;
  /** The content-length header. */
  protected int contentLength;

  protected Hashtable headers = new Hashtable();
  
  private int ServerPort = -1;
  
  
  
  /**
   * Create an httpResponse with the given response code with the given
   * payload. 
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payload The payload of the response.
   */
  protected httpResponse(int code, String contentType, BufferElement payload, Hashtable headertable) 
  {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = payload.size;
    this.headers = headertable;
    
    this.combinedData = null;
    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.header = new BufferElement(hdr);
    this.payload = payload;
  }
  
  
  /**
   * Create an httpResponse with the given response code with no payload.
   * A payload can be assigned later using setPayload().
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payload The payload of the response.
   */
  protected httpResponse(int code, String contentType, Hashtable headertable) 
  {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = 0; // Don't know it yet
    this.headers = headertable;
    
    //System.out.println("Generate response with code="+code);
    this.combinedData = null;
    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.header = new BufferElement(hdr);
    this.payload = null;
  }
  
  
  /**
   * Create an httpResponse with the given response code with the given
   * payload. 
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payload The payload of the response.
   */
  protected httpResponse(int code, String contentType, BufferElement payload) {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = payload.size;

    this.combinedData = null;
    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.header = new BufferElement(hdr);
    this.payload = payload;
  }

  /**
   * Create an httpResponse with the given response code with the given
   * payload. 
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payload The payload of the response.
   * @param contentLength The contentLength to place in the header.
   */
  protected httpResponse(int code, String contentType, BufferElement payload, int contentLength) 
  {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = contentLength;

    this.combinedData = null;
    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.header = new BufferElement(hdr);
    this.payload = payload;
  }

  /**
   * Create an httpResponse with the given response code with no payload.
   * A payload can be assigned later using setPayload().
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payload The payload of the response.
   */
  protected httpResponse(int code, String contentType) {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = 0; // Don't know it yet

    this.combinedData = null;
    this.header = null;
    this.payload = null;
  }

  /**
   * Create an httpResponse with the the given response code, with an
   * empty payload of the given size. This can be more efficient than
   * providing a payload separately, as the entire contents of the 
   * httpResponse can be sent as a single TCP packet. The payload can
   * be filled in using the getPayload() method.
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payloadSize The size of the payload to allocate.
   * @param compQ The completion queue for the payload.
   */
  protected httpResponse(int code, String contentType, int payloadSize, SinkIF compQ) 
  {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = payloadSize;

    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.combinedData = new BufferElement(hdr.length + payloadSize);
    combinedData.compQ = compQ;
    this.header = new BufferElement(combinedData.data, 0, hdr.length);
    System.arraycopy(hdr, 0, header.data, 0, hdr.length);
    this.payload = new BufferElement(combinedData.data, hdr.length, payloadSize);
  }

  /**
   * Create an httpResponse with the the given response code, with an
   * empty payload of the given size. This can be more efficient than
   * providing a payload separately, as the entire contents of the 
   * httpResponse can be sent as a single TCP packet. The payload can
   * be filled in using the getPayload() method.
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payloadSize The size of the payload to allocate.
   */
  protected httpResponse(int code, String contentType, int payloadSize) 
  {
    this(code, contentType, payloadSize, null);
  }

  /**
   * Create an httpResponse with the the given response code, with an
   * empty payload of the given size. This can be more efficient than
   * providing a payload separately, as the entire contents of the 
   * httpResponse can be sent as a single TCP packet. The payload can
   * be filled in using the getPayload() method.
   *
   * @param code The response code; should be one of the constants
   *  from httpResponse.RESPONSE_*.
   * @param contentType The MIME type of the response content. Should
   *  not be CRLF-terminated.
   * @param payloadSize The size of the payload to allocate.
   */
  protected httpResponse(int code, String contentType, int payloadSize, Hashtable headertable, boolean withheaders) 
  {
    this.code = code;
    this.contentType = contentType;
    this.contentLength = payloadSize;
    this.headers = headertable;
    
    String hdrString = genHeader();
    byte hdr[] = hdrString.getBytes();
    this.combinedData = new BufferElement(hdr.length + payloadSize);
    combinedData.compQ = null;
    this.header = new BufferElement(combinedData.data, 0, hdr.length);
    System.arraycopy(hdr, 0, header.data, 0, hdr.length);
    this.payload = new BufferElement(combinedData.data, hdr.length, payloadSize);
  }
  
  
  public void setServerPort(int val)
  {
    ServerPort = val;
  }
  
  public int getServerPort()
  {
    return ServerPort;   
  }
  
  /** 
   * Return the entity header as a String. Must be implemented by 
   * subclasses of httpResponse.
   */
  protected abstract String getEntityHeader();

  /**
   * Used to set the payload after creating the response with an 
   * empty payload. XXX Should not be used if the payload was allocated 
   * by this response (that is, if the payloadSize was specified in the 
   * constructor). 
   */
  public void setPayload(BufferElement payload) {
    this.payload = payload;
    this.contentLength = payload.size;
  }

  /**
   * Returns the header for this response.
   */
  public BufferElement getHeader() 
  {
    if (this.header == null) 
    {
      String hdrString = genHeader();
      byte hdr[] = hdrString.getBytes();
      this.header = new BufferElement(hdr);
    }
    return this.header;
  }

  /**
   * Returns the payload for this response.
   */
  public BufferElement getPayload() {
    return payload;
  }

  
  /**
   * Set the default header string sent in all responses.
   */
  public static void setDefaultHeader(String defhdr) 
  {
    defaultHeader = defhdr;
  }

  /**
   * Return the default header string sent in all responses.
   */
  public static String getDefaultHeader() {
    return defaultHeader;
  }

  /**
   * Generate the header.
   */
  private String genHeader() 
  {
    String hdrString;
    switch (code) 
    {
      case RESPONSE_CONTINUE :
        hdrString = HTTP_VERSION+" 100 CONTINUE\n"; break; 
      case RESPONSE_SWITCHING_PROTOCOLS :
        hdrString = HTTP_VERSION+" 101 SWITCHING PROTOCOLS\n"; break; 
        
      case RESPONSE_OK: 
	hdrString = HTTP_VERSION+" 200 OK\n"; break;
      case RESPONSE_CREATED:
	hdrString = HTTP_VERSION+" 201 CREATED\n"; break;
      case RESPONSE_ACCEPTED:
	hdrString = HTTP_VERSION+" 202 ACCEPTED\n"; break;  
      case RESPONSE_NON_AUTHORITATIVE_INFORMATION: 
	hdrString = HTTP_VERSION+" 203 Non-Authoritative Information\n"; break;
      case RESPONSE_NO_CONTENT:
	hdrString = HTTP_VERSION+" 204 NO CONTENT\n"; break;  
      case RESPONSE_RESET_CONTENT: 
	hdrString = HTTP_VERSION+" 205 RESET CONTENT\n"; break;
      case RESPONSE_PARTIAL_CONTENT:
	hdrString = HTTP_VERSION+" 206 PARTIAL CONTENT\n"; break;    
        
      case RESPONSE_MULTIPLE_CHOICES:
	hdrString = HTTP_VERSION+" 300 Multiple Choices\n"; break;  
      case RESPONSE_REDIRECT:
	hdrString = HTTP_VERSION+" 301 MOVED PERMANENTLY\n"; break;
      case RESPONSE_FOUND:
	hdrString = HTTP_VERSION+" 302 FOUND\n"; break;  
      case RESPONSE_SEE_OTHER:
	hdrString = HTTP_VERSION+" 303 SEE OTHER\n"; break;
      case RESPONSE_NOT_MODIFIED:
	hdrString = HTTP_VERSION+" 304 NOT MODIFIED\n"; break;  
        
      case RESPONSE_USE_PROXY:
	hdrString = HTTP_VERSION+" 305 USE PROXY\n"; break;  
      case RESPONSE_RESERVED_306:
	hdrString = HTTP_VERSION+" 306 Reserved\n"; break;
      case RESPONSE_TEMPORARY_REDIRECT:
	hdrString = HTTP_VERSION+" 307 TEMPORARY REDIRECT\n"; break;   
        
        
      case RESPONSE_BAD_REQUEST: 
	hdrString = HTTP_VERSION+" 400 BAD REQUEST\n"; break;
      case RESPONSE_UNAUTHORIZED: 
	hdrString = HTTP_VERSION+" 401 UNAUTHORIZED\n"; break;
      case RESPONSE_PAYMENT_REQUIRED: 
	hdrString = HTTP_VERSION+" 402 PAYMENT REQUIRED\n"; break;
      case RESPONSE_FORBIDDEN: 
	hdrString = HTTP_VERSION+" 403 FORBIDDEN\n"; break;
      case RESPONSE_NOT_FOUND:
	hdrString = HTTP_VERSION+" 404 NOT FOUND\n"; break; 
      case RESPONSE_METHOD_NOT_ALLOWED: 
	hdrString = HTTP_VERSION+" 405 Method Not Allowed\n"; break;
      case RESPONSE_NOT_ACCEPTABLE: 
	hdrString = HTTP_VERSION+" 406 Not Acceptable\n"; break;
      case RESPONSE_PROXY_AUTHENTICATION_REQUIRED: 
	hdrString = HTTP_VERSION+" 407 Proxy Authentication Required\n"; break;
      case RESPONSE_REQUEST_TIMEOUT: 
	hdrString = HTTP_VERSION+" 408 REQUEST TIMEOUT\n"; break;
      case RESPONSE_CONFLICT:
	hdrString = HTTP_VERSION+" 409 CONFLICT\n"; break;  
        
        
      case RESPONSE_GONE: 
	hdrString = HTTP_VERSION+" 410 GONE\n"; break;
      case RESPONSE_LENGTH_REQUIRED: 
	hdrString = HTTP_VERSION+" 411 LENGTH REQUIRED\n"; break;
      case RESPONSE_PRECONDITION_FAILED: 
	hdrString = HTTP_VERSION+" 412 PRECONDITION FAILED\n"; break;
      case RESPONSE_REQUEST_ENTITY_TOO_LARGE: 
	hdrString = HTTP_VERSION+" 413 Request Entity Too Large\n"; break;
      case RESPONSE_REQUEST_URI_TOO_LONG:
	hdrString = HTTP_VERSION+" 414 Request-URI Too Long\n"; break; 
      case RESPONSE_UNSUPPORTED_MEDIA_TYPE: 
	hdrString = HTTP_VERSION+" 415 Unsupported Media Type\n"; break;
      case RESPONSE_REQUESTED_RANGE_NOT_SATISFIABLE: 
	hdrString = HTTP_VERSION+" 416 Requested Range Not Satisfiable\n"; break;
      case RESPONSE_EXPECTATION_FAILED: 
	hdrString = HTTP_VERSION+" 417 Expectation Failed\n"; break;
        
        
      case RESPONSE_INTERNAL_SERVER_ERROR:
      	hdrString = HTTP_VERSION+" 500 INTERNAL SERVER ERROR\n"; break;
      case RESPONSE_NOT_IMPLEMENTED:
      	hdrString = HTTP_VERSION+" 501 Not Implemented\n"; break;
      case RESPONSE_BAD_GATEWAY:
      	hdrString = HTTP_VERSION+" 502 Bad Gateway\n"; break;
      case RESPONSE_SERVICE_UNAVAILABLE:
      	hdrString = HTTP_VERSION+" 503 SERVICE UNAVAILABLE\n"; break;
      case RESPONSE_GATEWAY_TIMEOUT:
      	hdrString = HTTP_VERSION+" 504 Gateway Timeout\n"; break;
      case RESPONSE_HTTP_VERSION_NOT_SUPPORTED:
      	hdrString = HTTP_VERSION+" 505 HTTP Version Not Supported\n"; break;  
        
        
      default: 
        System.out.println("Bad code in httpResponse: "+code);
        hdrString = HTTP_VERSION+" "+code+" UNKNOWN RESPONSE CODE\n"; break;
	//throw new Error("Bad code in httpResponse: "+code);
    } 
    if (defaultHeader != null) 
        hdrString += defaultHeader;
    if (contentType != null) 
    {
      if (!contentType.equalsIgnoreCase("null"))
      {
        if (!headers.containsKey("Content-Type"))  
          hdrString += "Content-Type: "+contentType+CRLF;
      }
    }
    if (contentLength != 0) 
    {
      if (!headers.containsKey("Content-Length"))      
        hdrString += "Content-Length: "+contentLength+CRLF;
    }
       
    
    //hdrString += "Connection: close"+CRLF;
            
    Enumeration enum = headers.keys();
    while (enum.hasMoreElements())
    {
      String hkey = (String) enum.nextElement();
      String hval = (String) headers.get(hkey);
      hdrString += hkey+": "+hval+CRLF;
    }
    String ehdr = getEntityHeader();
    if (ehdr != null) {
      hdrString += ehdr;
    }
    hdrString += CRLF;
    
    //System.out.println("headers in response \n\n"+hdrString+"\n");
    return hdrString;
  }
  
  /**
   * Get an array of BufferElements corresponding to this response.
   * Used internally when sending the response to a client.
   *
   * @param sendHeader Indicate whether the header should be included.
   */
  public BufferElement[] getBuffers(boolean sendHeader) {
    if (DEBUG) System.err.println("httpResponse: getBuffers() called");

    BufferElement bufarr[] = null;
    if (combinedData != null) {
      if (sendHeader) {
	if (DEBUG) System.err.println("httpResponse: Returning combinedData (len="+combinedData.size+")");
	bufarr = new BufferElement[1];
	bufarr[0] = combinedData;
      } else {
	if (DEBUG) System.err.println("httpResponse: Returning combinedData payload only (len="+payload.size+")");
	bufarr = new BufferElement[1];
	bufarr[0] = payload;
      }
    } else if (sendHeader) {
      if (payload != null) {
	if (DEBUG) System.err.println("httpResponse: Returning header and payload (paylen="+payload.size+")");
	bufarr = new BufferElement[2];
	bufarr[0] = getHeader();
	bufarr[1] = getPayload();
      } else {
	if (DEBUG) System.err.println("httpResponse: Returning header only (len="+header.size+")");
	bufarr = new BufferElement[1];
	bufarr[0] = getHeader();
      }
    } else {
      // Don't send header
      if (payload != null) {
	if (DEBUG) System.err.println("httpResponse: Returning payload only (paylen="+payload.size+")");
	bufarr = new BufferElement[1];
	bufarr[0] = payload;
      } else {
	if (DEBUG) System.err.println("httpResponse: Nothing to return!");
	bufarr = null;
      }
    }
      
    return bufarr;
  }

}
